# -*- coding: utf-8 -*-
# flake8: noqa
from __future__ import absolute_import, division, print_function, unicode_literals

import utool as ut
ut.noinject(__name__, '[dtool_ibeis]')

from dtool_ibeis import __SQLITE__ as lite
from dtool_ibeis import base
from dtool_ibeis import sql_control
from dtool_ibeis import depcache_control
from dtool_ibeis import depcache_table

from dtool_ibeis.depcache_control import DependencyCache, make_depcache_decors
from dtool_ibeis.base import (AlgoResult, MatchResult, Config,
                        VsManySimilarityRequest, VsOneSimilarityRequest)
from dtool_ibeis.depcache_table import ExternalStorageException, ExternType
from dtool_ibeis.base import *  # NOQA
from dtool_ibeis.sql_control import SQLDatabaseController

__version__ = '1.0.0'



# -*- coding: utf-8 -*-
"""
custom sqlite3 module that supports numpy types
"""
from __future__ import absolute_import, division, print_function
import sys
import six
import io
import uuid
import numpy as np
import utool as ut
from six.moves import input
ut.noinject(__name__, '[dtool_ibeis.__SQLITE__]')


VERBOSE_SQL = '--veryverbose' in sys.argv or '--verbose' in sys.argv or '--verbsql' in sys.argv
TRY_NEW_SQLITE3 = False

# SQL This should be the only file which imports sqlite3
if not TRY_NEW_SQLITE3:
    from sqlite3 import Binary, register_adapter, register_converter
    from sqlite3 import *  # NOQA

#try:
#    # Try to import the correct version of sqlite3
#    if VERBOSE_SQL:
#        from pysqlite2 import dbapi2
#        import sqlite3
#        print('dbapi2.sqlite_version  = %r' % dbapi2.sqlite_version)
#        print('sqlite3.sqlite_version = %r' % sqlite3.sqlite_version)
#        print('using dbapi2 as lite')
#        # Clean namespace
#        del sqlite3
#        del dbapi2
#    if not TRY_NEW_SQLITE3:
#        raise ImportError('user wants python sqlite3')
#    from pysqlite2.dbapi2 import *  # NOQA
#except ImportError as ex:
#    if VERBOSE_SQL:
#        print(ex)
#    # Fallback
#    from sqlite3 import *  # NOQA
#    if VERBOSE_SQL:
#        print('using sqlite3 as lite')


def REGISTER_SQLITE3_TYPES():
    def _read_numpy_from_sqlite3(blob):
        # INVESTIGATE: Is memory freed up correctly here?
        out = io.BytesIO(blob)
        out.seek(0)
        #return np.load(out)
        # Is this better?
        arr = np.load(out)
        out.close()
        return arr

    def _read_bool(b):
        return None if b is None else bool(b)

    def _write_bool(b):
        return b

    if six.PY2:
        def _write_numpy_to_sqlite3(arr):
            out = io.BytesIO()
            np.save(out, arr)
            out.seek(0)
            #return buffer(out.read())
            return Binary(out.read())
    else:
        def _write_numpy_to_sqlite3(arr):
            out = io.BytesIO()
            np.save(out, arr)
            out.seek(0)
            return memoryview(out.read())

    def _read_uuid_from_sqlite3(blob):
        try:
            return uuid.UUID(bytes_le=blob)
        except ValueError as ex:
            ut.printex(ex, keys=['blob'])
            raise
            print('WARNING: COULD NOT PARSE UUID %r, GIVING RANDOM' % (blob, ))
            input('continue... [enter]')
            return uuid.uuid4()

    if six.PY2:
        def _write_uuid_to_sqlite3(uuid_):
            #return buffer(uuid_.bytes_le)
            return Binary(uuid_.bytes_le)
    elif six.PY3:
        def _write_uuid_to_sqlite3(uuid_):
            return memoryview(uuid_.bytes_le)

    def register_numpy_dtypes():
        if VERBOSE_SQL:
            print('Register NUMPY dtypes with SQLite3')

        py_int_type = long if six.PY2 else int
        for dtype in (np.int8, np.int16, np.int32, np.int64,
                      np.uint8, np.uint16, np.uint32, np.uint64):
            register_adapter(dtype, py_int_type)
        register_adapter(np.float32, float)
        register_adapter(np.float64, float)

    def _read_dict_from_sqlite3(blob):
        return ut.from_json(blob)
        #return uuid.UUID(bytes_le=blob)
    def _write_dict_to_sqlite3(dict_):
        return ut.to_json(dict_)

    def register_numpy():
        """
        Tell SQL how to deal with numpy arrays
        Utility function allowing numpy arrays to be stored as raw blob data
        """
        if VERBOSE_SQL:
            print('Register NUMPY with SQLite3')
        register_converter('NUMPY', _read_numpy_from_sqlite3)
        register_converter('NDARRAY', _read_numpy_from_sqlite3)
        register_adapter(np.ndarray, _write_numpy_to_sqlite3)

    def register_uuid():
        """ Utility function allowing uuids to be stored in sqlite """
        if VERBOSE_SQL:
            print('Register UUID with SQLite3')
        register_converter('UUID', _read_uuid_from_sqlite3)
        register_adapter(uuid.UUID, _write_uuid_to_sqlite3)

    def register_dict():
        if VERBOSE_SQL:
            print('Register DICT with SQLite3')
        register_converter('DICT', _read_dict_from_sqlite3)
        register_adapter(dict, _write_dict_to_sqlite3)

    def register_list():
        if VERBOSE_SQL:
            print('Register LIST with SQLite3')
        register_converter('LIST', ut.from_json)
        register_adapter(list, ut.to_json)

    def register_bool():
        # FIXME: ensure this works
        if VERBOSE_SQL:
            print('Register BOOL with SQLite3')
        register_converter('BOOL', _read_bool)
        register_adapter(bool, _write_bool)

    register_numpy_dtypes()
    register_numpy()
    register_uuid()
    register_dict()
    register_list()
    # register_bool()  # TODO
REGISTER_SQLITE3_TYPES()


#def connect2(fpath, text_factory=None):
#    """ wrapper around lite.connect """
#    connection = connect(fpath, detect_types=PARSE_DECLTYPES)
#    return connection
#    #timeout=5,
#    # check_same_thread=False)
#    # isolation_level='DEFERRED',
#    # cached_statements=1000

TYPE_TO_SQLTYPE = {
    np.ndarray: 'NDARRAY',
    uuid.UUID: 'UUID',
    np.float32: 'REAL',
    np.float64: 'REAL',
    float: 'REAL',
    int: 'INTEGER',
    str: 'TEXT',
    # bool: 'BOOL',  # TODO
    bool: 'INTEGER',
    dict: 'DICT',
    list: 'LIST',
}

if six.PY2:
    TYPE_TO_SQLTYPE[six.text_type] = 'TEXT'

# Clean namespace
del REGISTER_SQLITE3_TYPES



    def _add_dirty_rows(table, parent_rowids, config_rowid, isdirty_list,
                        config, verbose=True):
        """ Does work of adding dirty rowids """
        dirty_parent_rowids = ut.compress(parent_rowids, isdirty_list)
        try:
            # CALL EXTERNAL PREPROCESSING / GENERATION FUNCTION
            if table.isalgo:
                # TODO: DEPRICATE OLD ALGO REQUST STRUCTURE
                # HACK: config here is a request
                request = config
                #subreq = request.shallow_copy # TODO
                # FIXME: Need to vsone querys and name-vs-name queries work
                # here.
                if table.productinput:
                    # Roundabout way of forcing algo requests into the depcache
                    # structure Very ugly
                    subreq_list = list(request.shallowcopy_vsonehack(
                        qmask=isdirty_list))
                    proptup_gen_list = [table.preproc_func(table.depc, subreq)
                                        for subreq in subreq_list]
                    from itertools import chain
                    proptup_gen = chain(*proptup_gen_list)
                    dirty_params_iter = table._yeild_algo_result(
                        dirty_parent_rowids, proptup_gen, config_rowid)
                    #proptup_gen = list(proptup_gen)
                else:
                    subreq = request.shallowcopy(qmask=isdirty_list)
                    # CALL REGISTRED ALGO WORKER FUNCTION
                    proptup_gen = table.preproc_func(table.depc, subreq)
                    dirty_params_iter = table._yeild_algo_result(
                        dirty_parent_rowids, proptup_gen, config_rowid)
            else:
                args = zip(*dirty_parent_rowids)
                if table._asobject:
                    # Convinience
                    args = [table.depc.get_obj(parent, rowids)
                            for parent, rowids in zip(table.parents, args)]
                # hack config out of request
                config_ = config.config if hasattr(config, 'config') else config
                # CALL REGISTRED TABLE WORKER FUNCTION
                proptup_gen = table.preproc_func(table.depc, *args,
                                                 config=config_)
                if len(table._nested_idxs) > 0:
                    assert not table.isalgo
                    unnest_data = table._make_unnester()
                    proptup_gen = (unnest_data(data) for data in proptup_gen)
                dirty_params_iter = table._concat_rowids_data(
                    dirty_parent_rowids, proptup_gen, config_rowid)

            chunksize = (len(dirty_parent_rowids)
                         if table.chunksize is None else table.chunksize)

            # TODO: Separate this as a function which can be specified as a
            # callback.
            num_chunks = int(ceil(len(dirty_parent_rowids) / chunksize))
            chunk_iter = ut.ichunks(dirty_params_iter, chunksize=chunksize)
            lbl = 'adding %s chunk' % (table.tablename)
            prog_iter = ut.ProgIter(chunk_iter, nTotal=num_chunks, lbl=lbl)
            for dirty_params_chunk in prog_iter:
                nInput = len(dirty_params_chunk)
                if table.isalgo:
                    # HACKS, really this should be for anything that has a
                    # extern write function
                    sql_chunks = table._save_algo_result(dirty_params_chunk)
                    table.db._add(table.tablename, table._table_colnames,
                                  sql_chunks, nInput=nInput)
                else:
                    table.db._add(table.tablename, table._table_colnames,
                                  dirty_params_chunk, nInput=nInput)
        except Exception as ex:
            ut.printex(ex, 'error in add_rowids', keys=[
                'table', 'table.parents', 'parent_rowids', 'config', 'args',
                'config_rowid', 'dirty_parent_rowids', 'table.preproc_func'])
            raise


    def _concat_rowids_data(table, dirty_parent_rowids, proptup_gen,
                            config_rowid):
        for parent_rowids, data_cols in zip(dirty_parent_rowids, proptup_gen):
            try:
                yield parent_rowids + (config_rowid,) + data_cols
            except Exception as ex:
                ut.printex(ex, 'cat error', keys=[
                    'config_rowid', 'data_cols', 'parent_rowids'])
                raise

    def _yeild_algo_result(table, dirty_parent_rowids, proptup_gen, config_rowid):
        # TODO: generalize to all external data that needs to be written
        # explicitly
        extern_fname_list = table._get_extern_fnames(dirty_parent_rowids,
                                                     config_rowid)
        extern_dpath = table._get_extern_dpath()
        ut.ensuredir(extern_dpath, verbose=True or table.depc._debug)
        fpath_list = [join(extern_dpath, fname) for fname in extern_fname_list]
        _iter = zip(dirty_parent_rowids, proptup_gen, fpath_list)
        for parent_rowids, algo_result, extern_fpath in _iter:
            yield parent_rowids, config_rowid, algo_result, extern_fpath

    def _save_algo_result(table, dirty_params_chunk):
        for tup in dirty_params_chunk:
            parent_rowids, config_rowid, algo_result, extern_fpath = tup
            try:
                algo_result.save_to_fpath(extern_fpath, True)
                yield parent_rowids + (config_rowid,) + (extern_fpath,)
            except Exception as ex:
                ut.printex(ex, 'cat2 error', keys=[
                    'config_rowid', 'data_cols', 'parent_rowids'])
                raise


class AlgoRequest(BaseRequest, ut.NiceRepr):
    """
    Base class for algo request objects
    Need this for TestResult Integration

    This class might not be need, and is being added for
    compatibility support.
    The problem it solve is having daids as part of a config.  A config should
    be used to specify algorithm parameters, but a referense set of matchable
    annotations seems to go beyond that.  Therefore, AlgoRequest.

    Ignore:
        cls = dtool_ibeis.AlgoRequest

    Example:
        >>> # ENABLE_DOCTEST
        >>> from dtool_ibeis.base import *  # NOQA
        >>> from dtool_ibeis.example_depcache import testdata_depc
        >>> depc = testdata_depc()
        >>> #request1 = depc.new_algo_request('vsone', [1, 2], [1, 2])
        >>> request2 = depc.new_request('vsmany', [1, 2], [1, 2])
    """
    _isnewreq = True
    _qaids_independent = True
    _daids_independent = False

    @classmethod
    def new_algo_request(cls, depc, algoname, qaids, daids, cfgdict=None):
        request = cls()
        request._qaids = None
        request._daids = None

        request.depc = depc
        request.qaids = qaids
        request.daids = daids
        if cfgdict is None:
            cfgdict = {}
        configclass = depc.configclass_dict[algoname]
        config = configclass(**cfgdict)

        request.config = config
        request.algoname = algoname

        # hack
        request.params = dict(config.parse_items())
        return request

    @property
    def ibs(request):
        """ HACK specific to ibeis """
        if request.depc is None:
            return None
        return request.depc.controller

    def get_external_data_config2(request):
        # HACK
        #return None
        #print('[d] request.params = %r' % (request.params,))
        return request.params

    def get_external_query_config2(request):
        # HACK
        #return None
        #print('[q] request.params = %r' % (request.params,))
        return request.params

    @property
    def qaids(request):
        return request._qaids

    @qaids.setter
    def qaids(request, qaids):
        request._qaids = safeop(np.array, qaids)

    @property
    def daids(request):
        return request._daids

    @property
    def cfgstr(request):
        return request.get_cfgstr()

    @daids.setter
    def daids(request, daids):
        request._daids = safeop(np.array, daids)

    def get_parent_rowids(request):
        if request._daids_independent:
            parent_rowids = list(product(request.qaids, request.daids))
        else:
            parent_rowids = list(zip(request.qaids))
        return parent_rowids

    def execute(request, qaids=None, use_cache=None):
        if qaids is not None:
            qaids = [qaids] if not ut.isiterable(qaids) else qaids
            subreq = request.shallowcopy(qaids=qaids)
            return subreq.execute(use_cache=True)
        else:
            tablename = request.algoname
            table = request.depc[tablename]
            if use_cache is None:
                use_cache = not ut.get_argflag('--nocache')

            parent_rowids = request.get_parent_rowids()
            rowids = table.get_rowid(parent_rowids, config=request,
                                     recompute=not use_cache)
            result_list = table.get_row_data(rowids)
            return ut.get_list_column(result_list, 0)

    def shallowcopy_vsonehack(request, qmask=None, qaids=None):
        # Roundabout way of forcing algo requests into the depcache structure
        # Very ugly
        parent_rowids = request.get_parent_rowids()
        dirty_parents = ut.compress(parent_rowids, qmask)
        dirty_qaids = ut.take_column(dirty_parents, 0)
        dirty_daids = ut.take_column(dirty_parents, 1)
        groupxs = ut.group_indices(dirty_qaids)[1]
        daids_list = ut.apply_grouping(dirty_daids, groupxs)
        qaids_list = ut.apply_grouping(dirty_qaids, groupxs)
        for qaids, daids in zip(qaids_list, daids_list):
            #subreq = copy.copy(request)  # copy calls setstate and getstate
            subreq = request.__class__()
            subreq.__dict__.update(request.__dict__)
            subreq.qaids = qaids
            subreq.qaids = daids
            yield subreq

    def shallowcopy(request, qmask=None, qaids=None):
        """
        Creates a copy of qreq with the same qparams object and a subset of the
        qx and dx objects.  used to generate chunks of vsone and vsmany queries
        """
        #subreq = copy.copy(request)  # copy calls setstate and getstate
        subreq = request.__class__()
        subreq.__dict__.update(request.__dict__)
        if qmask is not None:
            assert qaids is None, 'cannot specify both'
            qaid_list  = subreq.qaids
            subreq.qaids = ut.compress(qaid_list, qmask)
        elif qaids is not None:
            subreq.qaids = qaids
        return subreq

    def get_query_hashid(request):
        return request._get_rootset_hashid(request.qaids, 'Q')

    def get_data_hashid(request):
        return request._get_rootset_hashid(request.daids, 'D')

    def get_pipe_cfgstr(request):
        return request.config.get_cfgstr()

    def get_pipe_hashid(request):
        return ut.hashstr27(request.get_pipe_cfgstr())

    def get_cfgstr(request, with_input=None, with_data=None, with_pipe=True,
                   hash_pipe=False):
        r"""
        main cfgstring used to identify the 'querytype'
        """
        if with_input is None:
            #with_input = False
            with_input = not request._qaids_independent

        if with_data is None:
            #with_data = True
            # non-independent aids must be in config string
            with_data = not request._daids_independent

        cfgstr_list = []
        if with_input:
            cfgstr_list.append(request.get_query_hashid())
        if with_data:
            cfgstr_list.append(request.get_data_hashid())
        if with_pipe:
            if hash_pipe:
                cfgstr_list.append(request.get_pipe_hashid())
            else:
                cfgstr_list.append(request.get_pipe_cfgstr())
        cfgstr = '_'.join(cfgstr_list)
        return cfgstr

    def get_full_cfgstr(request):
        """ main cfgstring used to identify the algo hash id """
        full_cfgstr = request.get_cfgstr(with_input=True)
        return full_cfgstr

    def __nice__(request):
        dbname = (None if request.depc is None or request.depc.controller is None
                  else request.depc.controller.get_dbname())
        infostr_ = 'nQ=%s, nD=%s %s' % (len(request.qaids), len(request.daids),
                                        request.get_pipe_hashid())
        return '(%s) %s' % (dbname, infostr_)

    #def _get_rootset_hashid(request, root_rowids, prefix):
    #    uuid_type = 'V'
    #    label = ''.join((prefix, uuid_type, 'UUIDS'))
    #    uuid_list = request.depc.get_root_uuid(root_rowids)
    #    #uuid_hashid = ut.hashstr_arr27(uuid_list, label, pathsafe=True)
    #    uuid_hashid = ut.hashstr_arr27(uuid_list, label, pathsafe=False)
    #    return uuid_hashid

    #def __getstate__(request):
    #    state_dict = request.__dict__.copy()
    #    # SUPER HACK
    #    state_dict['dbdir'] = request.depc.controller.get_dbdir()
    #    del state_dict['depc']
    #    del state_dict['config']
    #    return state_dict

    def __setstate__(request, state_dict):
        import ibeis
        dbdir = state_dict['dbdir']
        del state_dict['dbdir']
        params = state_dict['params']
        depc = ibeis.opendb(dbdir=dbdir, web=False).depc
        configclass = depc.configclass_dict[state_dict['algoname'] ]
        config = configclass(**params)
        state_dict['depc'] = depc
        state_dict['config'] = config
        request.__dict__.update(state_dict)



    #def _row_exists(table, parent_rowids, config=None, eager=True, nInput=None,
    #                _debug=None):
    #    config_rowid = table.get_config_rowid(config=config)
    #    andwhere_colnames = table.superkey_colnames
    #    params_iter = (rowids + (config_rowid,) for rowids in parent_rowids)
    #    params_iter = list(params_iter)
    #    tblname = table.tablename
    #    flag_list = table.db.exists_where2(tblname, params_iter,
    #                                       andwhere_colnames, eager=eager,
    #                                       nInput=nInput)
    #    return flag_list


        path2_pidx = ut.make_index_lookup(path_edges_nodata, dict_factory=ut.odict)
        assert isinstance(path2_pidx, ut.odict)
        # Build mapping from each edge to the paths that is a part of
        pidx_list = ut.flatten([[idx] * len(path) for path, idx in path2_pidx.items()])
        edge_list = ut.flatten(accum_path_edges)
        edge_nodata_list = ut.flatten(path_edges_nodata)
        edge_nodata_to_datas = ut.group_items(edge_list, edge_nodata_list)
        # edge_hashable_list = [repr(edge) for edge in edge_list]
        # unique_edge_flags1 = ut.flag_unique_items(edge_nodata_list)
        # unique_edge_flags2 = ut.flag_unique_items(edge_hashable_list)
        # assert unique_edge_flags2 == unique_edge_flags1
        edge2_pidx = dict(ut.group_items(pidx_list, edge_nodata_list))
        # unique_edges = list(set(edge_nodata_list))

        type_to_paths = ut.ddict(list)
        for edge_nodata, edges in edge_nodata_to_datas.items():
            print('edge = %r' % (edge_nodata,))
            u, v, k = edge_nodata
            rinput_path_ids = ut.take_column(ut.take_column(edges, 3), 'rinput_path_id')
            rinput_path_id = rinput_path_ids[0]
            print('rinput_path_ids = %r' % (rinput_path_ids,))
            # edge_data = graph.edge[u][v][k]
            # local_input_id = edge_data['local_input_id']
            # print('local_input_id = %r' % (local_input_id,))
            pidxs = edge2_pidx[edge_nodata]
            paths = ut.take(accum_path_edges, pidxs)
            # Only take the path up to the current edge?
            # paths = [path[:path.index(edge) + 1] for path in paths]
            # print('paths = %s' % ut.repr3(paths, nl=2))
            # print('--------')
            type_to_paths[rinput_path_id].extend(paths)
            # type_to_paths[local_input_id].extend(paths)
            # type_to_pidxs[local_input_id] = pidxs
            # edge_type = edge_data['edge_type']
            # if edge_type != 'normal':
            #     pidxs = edge2_pidx[(u, v, k)]
            #     type_to_pidxs[edge_type] = pidxs
            #     type_to_paths[edge_type].extend(ut.take(path_edges, pidxs))
        # type_to_pidxs = {}

        expanded_input_graph = {}
        for type_, paths in type_to_paths.items():
            sub_edges = ut.flatten(paths)
            subgraph = ut.subgraph_from_edges(graph, sub_edges)
            expanded_input_graph[type_] = subgraph

        # use_normal = False
        # if use_normal:
        #     normal_pidxs = ut.index_complement(
        #         ut.flatten(type_to_pidxs.values()), len(path_edges))
        #     paths = ut.take(path_edges, normal_pidxs)
        #     sub_edges = ut.flatten(paths)
        #     subgraph = ut.subgraph_from_edges(graph, sub_edges)
        #     if len(subgraph.node) > 0:
        #         expanded_input_graph['normal'] = subgraph

    @ut.memoize
    def nonfinal_compute_order(table):
        """
        Returns which nodes to compute first, and what inputs are needed

            >>> from dtool_ibeis.depcache_control import *  # NOQA
            >>> from dtool_ibeis.example_depcache import testdata_depc
            >>> import plottool_ibeis as pt
            >>> pt.ensureqt()
            >>> depc = testdata_depc()
            >>> tablename = 'neighbs'
            >>> tablename = 'multitest_score'
            >>> table = depc[tablename]
            >>> nonfinal_compute_order = table.nonfinal_compute_order()
            >>> print(ut.repr3(nonfinal_compute_order))

        """
        import networkx as nx
        expanded_input_graph = table.expanded_input_graph
        composed_graph = nx.compose_all(expanded_input_graph.values())
        #pt.show_nx(composed_graph)
        topsort = nx.topological_sort(composed_graph)
        type_to_dependlevels = ut.map_dict_vals(ut.level_order,
                                                expanded_input_graph)
        level_orders = type_to_dependlevels
        # Find computation order for all dependencies
        nonfinal_compute_order = ut.merge_level_order(level_orders, topsort)
        return nonfinal_compute_order

    @property
    @ut.memoize
    def expected_input_order(table):
        """
        Returns what input (to depc.get_rowids) ordering should be be in
        parent_rowids
        """
        from six.moves import zip_longest
        nonfinal_compute_order = table.nonfinal_compute_order()
        expanded_input_graph = table.expanded_input_graph
        hgroupids = ut.ddict(list)
        for _tablename, order in reversed(nonfinal_compute_order):
            if _tablename == table.depc.root:
                continue
            for t in order:
                s = expanded_input_graph[t]
                colxs = [y['parent_colx'] for x in s.pred[_tablename].values()
                         for y in x.values()]
                assert len(colxs) > 0
                colx = min(colxs)
                #order_colxs.append(colx)
                hgroupids[t].append(colx)

        hgroupids = dict(hgroupids)

        keys = hgroupids.keys()
        vals = hgroupids.values()
        groupids = list(zip_longest(*vals, fillvalue=0))
        hgroups = ut.hierarchical_group_items(keys, groupids)
        fgroups = ut.flatten_dict_items(hgroups)
        fkey_list = [int(''.join(map(str, key))) for key in fgroups.keys()]
        fval_list = fgroups.values()

        dupkeys = ut.find_duplicate_items(fkey_list)
        assert len(dupkeys) == 0, 'cannot have duplicate orderings'

        expected_input_order = ut.flatten(ut.sortedby(fval_list, fkey_list))
        return expected_input_order

        if False:
            nodes = ut.all_nodes_between(graph, source, target)
            tablegraph = graph.subgraph(nodes)
            import plottool_ibeis as pt
            # pt.show_nx(tablegraph.reverse())
            # sink = ut.nx_sink_nodes(tablegraph)[0]
            # bfs_edges = list(ut.bfs_multi_edges(G, sink, data=True, reverse=True))
            G = tablegraph
            source = ut.nx_source_nodes(tablegraph)[0]
            bfs_edges = list(ut.bfs_multi_edges(G, source, data=0, reverse=False))
            print('bfs_edges = %r' % (bfs_edges,))
            T = nx.MultiDiGraph()
            T.add_node(source)
            T.add_edges_from(bfs_edges)
            pt.show_nx(T)

            def find_suffix(k, d):
                suffix = ''
                if d['ismulti']:
                    suffix += '_SET'
                if k != 0:
                    suffix += '_X' + str(k)
                return suffix

            G2 = nx.MultiDiGraph()
            # for u, v, k, d in G.edges(keys=True, data=True):
            edge_iter = ((u, v, k, d) for u in nx.topological_sort(G)[::-1] for v, kd in G[u].items() for k, d in kd.items())
            edges = list(edge_iter)
            for u, v, k, d in edges:
                s0 = ''
                # s0 = '_X0'
                suffix = find_suffix(k, d)
                if len(suffix) == 0:
                    G2.add_edge(u + s0, v + s0, attr_dict=d)
                else:
                    G2.add_edge(u + suffix, v + s0, attr_dict=d)
                    path_list = list(ut.all_multi_paths(G, source, u, data=True))
                    for path in path_list:
                        rpath = ut.reverse_path_edges(path)
                        parent_suffix = suffix
                        for redge in rpath:
                            v2, u2, k2, d2 = redge
                            u2 += parent_suffix
                            parent_suffix += find_suffix(k2, d2)
                            v2 += parent_suffix
                            if not G2.has_edge(u2, v2):
                                # if p2 not in G2.node:
                                G2.add_edge(u2, v2)

            pt.show_nx(G2)
            pt.show_nx(G)

        def compress_rinput_pathid(rinput_path_id):
            prev = None  # rinput_path_id[0]
            compressed = []
            for item in rinput_path_id:
                #if item != prev and not (item == '1' and prev == '2'):
                #if item != prev:
                compressed.append(item)
                # else:
                #     compressed.append(prev)
                #prev = item
            if len(compressed) > 1:
                compressed = compressed[1:]
            compressed = tuple(compressed)
            return compressed
        [[edge[3]['rinput_path_id'] for edge in path] for path in accum_path_edges]
        x = [[compress_rinput_pathid(edge[3]['rinput_path_id']) for edge in path] for path in accum_path_edges]

            >>> for type_, subgraph in expanded_input_graph.items():
            >>>     inter.append_plot(ut.partial(pt.show_nx, subgraph,
            >>>                                  title=type_))
            >>> composed_graph = nx.compose_all(expanded_input_graph.values())
            >>> inter.append_plot(ut.partial(pt.show_nx, composed_graph,
            >>>                              title='composed'))
        path_edges_nodata = ut.lmap(tuple, ut.lmap(ut.take_column, accum_path_edges, colx=slice(0, 3)))


def get_all_ancestor_rowids(depc, tablename, native_rowids):
    r"""
    Gets the root_rowids of the root table associated with the
    `native_rowids` of `tablename.

    Args:
        tablename (str):
        root_rowids (list):
        config (None): (default = None)

    Returns:
        dict: rowid_dict

    CommandLine:
        python -m dtool_ibeis.depcache_control --exec-get_all_ancestor_rowids

    Example:
        >>> # ENABLE_DOCTEST
        >>> from dtool_ibeis.depcache_control import *  # NOQA
        >>> from dtool_ibeis.example_depcache import testdata_depc
        >>> depc = testdata_depc()
        >>> tablename = 'spam'
        >>> target_root_rowids = [4, 9, 7]
        >>> native_rowids = depc.get_rowids(tablename, target_root_rowids)
        >>> rowid_dict = depc.get_all_ancestor_rowids(tablename, native_rowids)
        >>> root_rowids = list(rowid_dict[depc.root])
        >>> print(ut.repr3(rowid_dict, nl=1))
        >>> assert root_rowids == target_root_rowids
    """
    if depc._debug:
        print('[depc.descendant] GET ANSCESTOR ROWIDS %s ' % (tablename,))
    dependency_levels = depc.get_dependencies(tablename)
    rowid_dict = {}
    rowid_dict[tablename] = native_rowids

    # FIXME: not implemented very efficiently
    # Can do shortest existing path instead

    for level_keys in dependency_levels[::-1]:
        for tablekey in level_keys:
            if tablekey == depc.root:
                break
            table = depc[tablekey]
            child_rowids = rowid_dict[tablekey]
            colnames = table.parent_id_colnames
            parent_rowids_listT = table.get_internal_columns(
                child_rowids, colnames, keepwrap=True)
            parent_rowids_list = list(zip(*parent_rowids_listT))
            for parent_key, parent_rowids in zip(table.parent, parent_rowids_list):
                rowid_dict[parent_key] = parent_rowids
    return rowid_dict

def get_ancestor_rowids(depc, tablename, native_rowids, ancestor_tablename=None):
    """
    ancestor_tablename = depc.root
    native_rowids = cid_list
    tablename = const.CHIP_TABLE
    """
    if ancestor_tablename is None:
        ancestor_tablename = depc.root
    # rowid_dict = depc.get_all_ancestor_rowids(tablename, native_rowids)
    # ancestor_rowids = list(rowid_dict[ancestor_tablename])
    table = depc[tablename]
    ancestor_rowids = table.get_ancestor_rowids(native_rowids, ancestor_tablename)
    return ancestor_rowids


#def _parse_sqlkw(kwargs):
#    default_sqlkw = dict(
#        _debug=None, ensure=True, recompute=False, recompute_all=False,
#        eager=True, nInput=None, read_extern=True, onthefly=False,
#    )
#    otherkw = kwargs.copy()
#    sqlkw = {key: otherkw.pop(key, val) for key, val in default_sqlkw.items()}
#    return sqlkw, otherkw


    def get_dependants(depc, tablename):
        """
        gets level dependences table to the leaves. ie ancestors

        Example:
            >>> # ENABLE_DOCTEST
            >>> from dtool_ibeis.depcache_control import *  # NOQA
            >>> from dtool_ibeis.example_depcache import testdata_depc
            >>> depc = testdata_depc()
            >>> tablename = 'chip'
            >>> result = ut.repr3(depc.get_dependants(tablename), nl=1)
            >>> print(result)

            [
                ['chip'],
                ['keypoint'],
                ['fgweight', 'nnindexer', 'descriptor'],
                ['spam'],
                ['multitest'],
                ['multitest_score'],
            ]
        """
        # get_descendant_levels
        edges = depc.get_edges()
        children_, parents_ = list(zip(*edges))
        parent_to_children = ut.group_items(parents_, children_)
        to_leafs = {tablename: ut.path_to_leafs(tablename, parent_to_children)}
        dependency_levels_ = ut.get_levels(to_leafs)
        dependency_levels = ut.longest_levels(dependency_levels_)
        return dependency_levels



# -*- coding: utf-8 -*-
"""
CommandLine:
    python -m dtool_ibeis.example_depcache --exec-dummy_example_depcacahe --show
    python -m dtool_ibeis.depcache_control --exec-make_graph --show
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import utool as ut
import numpy as np
import uuid
from os.path import join, dirname
from six.moves import zip
from dtool_ibeis import depcache_control
import dtool_ibeis


if False:
    # Example of global registration
    DUMMY_ROOT_TABLENAME = 'dummy_annot'
    _depcdecors = depcache_control.make_depcache_decors(DUMMY_ROOT_TABLENAME)
    register_preproc = _depcdecors['preproc']
    register_subprop = _depcdecors['subprop']

    @register_preproc(tablename='dummy', parents=[DUMMY_ROOT_TABLENAME],
                      colnames=['data'], coltypes=[str])
    def dummy_global_preproc_func(depc, parent_rowids, config=None):
        if config is None:
            config = {}
        print('Requesting global dummy ')
        for rowid in parent_rowids:
            yield 'dummy'


class DummyKptsConfig(dtool_ibeis.Config):
    def get_param_info_list(self):
        return [
            ut.ParamInfo('adapt_shape', True),
            ut.ParamInfo('adapt_angle', False),
        ]


class DummyIndexerConfig(dtool_ibeis.Config):
    _param_info_list = [
        ut.ParamInfo('index_method', 'single'),
        ut.ParamInfo('trees', 8),
        ut.ParamInfo('algorithm', 'kdtree'),
    ]
    # FIXME: triggers duplicate error
    #_sub_config_list = [
    #    DummyKptsConfig
    #]


class DummyNNConfig(dtool_ibeis.Config):
    def get_param_info_list(self):
        return [
            ut.ParamInfo('K', 4),
            ut.ParamInfo('Knorm', 1),
            ut.ParamInfo('checks', 800),
            ut.ParamInfo('version', 1),
        ]


class DummySVERConfig(dtool_ibeis.Config):
    _param_info_list = [
        ut.ParamInfo('sver_on', True),
        ut.ParamInfo('xy_thresh', .01)
    ]


class DummyChipConfig(dtool_ibeis.Config):
    """
    Example:
        >>> # ENABLE_DOCTEST
        >>> from dtool_ibeis.example_depcache import *  # NOQA
        >>> cfg = DummyChipConfig()
        >>> cfg.dim_size = 700
        >>> cfg.histeq = True
        >>> print(cfg)
        >>> cfg.histeq = False
        >>> print(cfg)
    """
    _param_info_list = [
        ut.ParamInfo('resize_dim', 'width',
                     valid_values=['area', 'width', 'heigh', 'diag']),
        ut.ParamInfo('dim_size', 500, 'sz'),
        ut.ParamInfo('preserve_aspect', True),
        ut.ParamInfo('histeq', False, hideif=False),
        ut.ParamInfo('ext', '.png'),
        ut.ParamInfo('version', 0),
    ]


class ProbchipConfig(dtool_ibeis.Config):
    """
    CommandLine:
        python -m dtool_ibeis.example_depcache --exec-ProbchipConfig --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from dtool_ibeis.depcache_control import *  # NOQA
        >>> from dtool_ibeis.example_depcache import testdata_depc
        >>> depc = testdata_depc()
        >>> table = depc['probchip']
        >>> exec(ut.execstr_funckw(table.get_rowid), globals())
        >>> config = table.configclass(testerror=True)
        >>> root_rowids = [1, 2, 3]
        >>> parent_rowids = list(zip(root_rowids))
        >>> proptup_gen = list(table.preproc_func(depc, root_rowids, config))
        >>> pc_rowids = depc.get_rowids('probchip', root_rowids, config)
        >>> prop_list2 = depc.get('probchip', root_rowids, config=config, read_extern=False)
        >>> print(prop_list2)
        >>> #depc.new_request('probchip', [1, 2, 3])
        >>> fg_rowids = depc.get_rowids('fgweight', root_rowids, config)
        >>> fg = depc.get('fgweight', root_rowids, config=config)
        >>> #############
        >>> config = table.configclass(testerror=False)
        >>> root_rowids = [1, 2, 3]
        >>> parent_rowids = list(zip(root_rowids))
        >>> proptup_gen = list(table.preproc_func(depc, root_rowids, config))
        >>> pc_rowids2 = depc.get_rowids('probchip', root_rowids, config)
        >>> prop_list2 = depc.get('probchip', root_rowids, config=config, read_extern=False)
        >>> print(prop_list2)
        >>> #depc.new_request('probchip', [1, 2, 3])
        >>> fg_rowids2 = depc.get_rowids('fgweight', root_rowids, config)
    """
    _param_info_list = [
        ut.ParamInfo('testerror', False, hideif=False),
        ut.ParamInfo('ext', '.png', hideif='.png'),
    ]


class DummyVsManyConfig(dtool_ibeis.Config):
    # Different pipeline components can go here as well as dependencies
    # that were not explicitly enumerated in the tree structure
    _param_info_list = [
        #ut.ParamInfo('score_method', 'csum'),
        # should this be the only thing here?
        #ut.ParamInfo('daids', None),
        ut.ParamInfo('distinctiveness_model', None),
        ut.ParamInfo('version', 2),
    ]
    _sub_config_list = [
        # I guess different annots might want different configs ...
        DummyChipConfig,
        DummyKptsConfig,
        DummyIndexerConfig,
        DummyNNConfig,
        DummySVERConfig
    ]


class DummyVsOneConfig(dtool_ibeis.Config):
    def get_sub_config_list(self):
        # Different pipeline components can go here
        # as well as dependencies that were not
        # explicitly enumerated in the tree structure
        return [
            # I guess different annots might want different configs ...
            DummyChipConfig,
            DummyKptsConfig,
            DummyIndexerConfig,
            DummyNNConfig,
            DummySVERConfig
        ]

    def get_param_info_list(self):
        return [
            ut.ParamInfo('distinctiveness_model', None),
            ut.ParamInfo('ratio_thresh', None),
        ]


class DummyVsOneRequest(dtool_ibeis.VsOneSimilarityRequest):
    pass


class DummyVsManyRequest(dtool_ibeis.VsManySimilarityRequest):
    """
    Example:
        >>> # ENABLE_DOCTEST
        >>> from dtool_ibeis.example_depcache import *  # NOQA
        >>> algo_config = DummyVsManyConfig()
        >>> print(algo_config)
    """
    pass


class DummyAnnotMatch(dtool_ibeis.MatchResult):
    pass


class DummyVsOneMatch(dtool_ibeis.AlgoResult, ut.NiceRepr):
    def __init__(self):
        self.score = None
        self.qaid = None
        self.daid = None
        self.fm = None

    def __nice__(self):
        return ('(%d-vs-%d) %.2f' % (self.qaid, self.daid, self.score))


def testdata_depc(fname=None):
    """
    Example of local registration
    """

    import dtool_ibeis
    import vtool_ibeis as vt
    gpath_list = ut.lmap(ut.grab_test_imgpath, ut.get_valid_test_imgkeys(),
                         verbose=False)

    dummy_root = 'dummy_annot'

    def get_root_uuid(aid_list):
        return ut.lmap(ut.hashable_to_uuid, aid_list)

    # put the test cache in the dtool_ibeis repo
    dtool_repo = dirname(ut.get_module_dir(dtool_ibeis))
    cache_dpath = join(dtool_repo, 'DEPCACHE')

    depc = dtool_ibeis.DependencyCache(
        root_tablename=dummy_root, default_fname=fname,
        cache_dpath=cache_dpath,
        get_root_uuid=get_root_uuid,
        #root_asobject=root_asobject,
        use_globals=False)

    @depc.register_preproc(tablename='chip', parents=[dummy_root],
                           colnames=['size', 'chip'],
                           coltypes=[(int, int), ('extern', vt.imread, vt.imwrite)],
                           configclass=DummyChipConfig)
    def dummy_preproc_chip(depc, annot_rowid_list, config=None):
        """
        TODO: Infer properties from docstr?

        Args:
            depc (dtool_ibeis.DependencyCache):
            annot_rowid_list (list): list of annot rowids
            config (dict): config dictionary

        Returns:
            tuple : ((int, int), ('extern', vt.imread))
        """
        if config is None:
            config = {}
        # Demonstates using asobject to get input to function as a dictionary
        # of properties
        #for annot in annot_list:
        #print('[preproc] Computing chips of aid=%r' % (aid,))
        print('[preproc] Computing chips')
        for aid in annot_rowid_list:
            #aid = annot['aid']
            #chip_fpath = annot['gpath']
            chip_fpath = gpath_list[aid]
            #w, h = vt.image.open_image_size(chip_fpath)
            chip = vt.imread(chip_fpath)
            size = vt.get_size(chip)
            #size = (w, h)
            print('Dummpy preproc chip yeilds')
            print('* chip_fpath = %r' % (chip_fpath,))
            print('* size = %r' % (size,))
            #yield size, chip_fpath
            yield size, chip

    @depc.register_preproc(
        'probchip', [dummy_root], ['size', 'probchip'],
        coltypes=[(int, int), ('extern', vt.imread, vt.imwrite, '.png')],
        configclass=ProbchipConfig,
    )
    def dummy_preproc_probchip(depc, root_rowids, config):
        print('[preproc] Computing probchip')
        for rowid in root_rowids:
            if config['testerror']:
                if rowid % 2 == 0:
                    # Test error yeilds None on even rowids
                    yield None
                    continue
            rng = np.random.RandomState(rowid)
            probchip = rng.randint(0, 255, size=(64, 64))
            #probchip = np.zeros((64, 64))
            size = (rowid, rowid)
            yield size, probchip

    @depc.register_preproc(
        'keypoint', ['chip'], ['kpts', 'num'], [np.ndarray, int],
        #default_onthefly=True,
        configclass=DummyKptsConfig,
        docstr='Used to store individual chip features (ellipses)',)
    def dummy_preproc_kpts(depc, chip_rowids, config=None):
        if config is None:
            config = {}
        print('config = %r' % (config,))
        adapt_shape = config['adapt_shape']
        print('[preproc] Computing kpts')

        ut.assert_all_not_None(chip_rowids, 'chip_rowids')
        # This is in here to attempt to trigger a failure of the chips dont
        # exist and the feature cache is called.
        chip_fpath_list = depc.get_native('chip', chip_rowids, 'chip',
                                          read_extern=False)
        print('computing featurse from chip_fpath_list = %r' % (chip_fpath_list,))

        for rowid in chip_rowids:
            if adapt_shape:
                kpts = np.zeros((7 + rowid, 6)) + rowid
            else:
                kpts = np.ones((7 + rowid, 6)) + rowid
            num = len(kpts)
            yield kpts, num

    @depc.register_preproc('descriptor', ['keypoint'], ['vecs'], [np.ndarray],)
    def dummy_preproc_vecs(depc, kp_rowid, config=None):
        if config is None:
            config = {}
        print('[preproc] Computing vecs')
        for rowid in kp_rowid:
            yield np.ones((7 + rowid, 8), dtype=np.uint8) + rowid,

    @depc.register_preproc('fgweight', ['keypoint', 'probchip'], ['fgweight'], [np.ndarray],)
    def dummy_preproc_fgweight(depc, kpts_rowid, probchip_rowid, config=None):
        if config is None:
            config = {}
        print('[preproc] Computing fgweight')
        for rowid1, rowid2 in zip(kpts_rowid, probchip_rowid):
            yield np.ones(7 + rowid1),

    @depc.register_preproc(
        tablename='vsmany', colnames='annotmatch', coltypes=DummyAnnotMatch,
        requestclass=DummyVsManyRequest, configclass=DummyVsManyConfig)
    def vsmany_matching(depc, qaids, config=None):
        """
        CommandLine:
            python -m dtool_ibeis.base --exec-VsManySimilarityRequest
        """
        print('RUNNING DUMMY VSMANY ALGO')
        daids = config.daids
        qaids = qaids

        sver_on = config.dummy_sver_cfg['sver_on']
        kpts_list = depc.get_property('keypoint', list(qaids))  # NOQA
        #dummy_preproc_kpts
        for qaid in qaids:
            dnid_list = [1, 1, 2, 2]
            unique_nids = [1, 2]
            if sver_on:
                annot_score_list = [.2, .2, .4, .5]
                name_score_list = [.2, .5]
            else:
                annot_score_list = [.3, .3, .6, .9]
                name_score_list = [.1, .7]
            annot_match = DummyAnnotMatch(qaid, daids, dnid_list,
                                          annot_score_list, unique_nids,
                                          name_score_list)
            yield annot_match

    SIMPLE = 0
    if not SIMPLE:

        @depc.register_preproc(
            tablename='chipmask', parents=[dummy_root], colnames=['size', 'mask'],
            coltypes=[(int, int), ('extern', vt.imread, vt.imwrite)])
        def dummy_manual_chipmask(depc, parent_rowids, config=None):
            import vtool_ibeis as vt
            from plottool_ibeis import interact_impaint
            mask_dpath = join(depc.cache_dpath, 'ManualChipMask')
            ut.ensuredir(mask_dpath)
            if config is None:
                config = {}
            print('Requesting user defined chip mask')
            for rowid in parent_rowids:
                img = vt.imread(gpath_list[rowid])
                mask = interact_impaint.impaint_mask2(img)
                mask_fpath = join(mask_dpath, 'mask%d.png' % (rowid,))
                vt.imwrite(mask_fpath, mask)
                w, h = vt.get_size(mask)
                yield (w, h), mask_fpath

        @depc.register_preproc('notch', [dummy_root], ['notchdata'], [np.ndarray],)
        def dummy_preproc_notch(depc, parent_rowids, config=None):
            if config is None:
                config = {}
            print('[preproc] Computing notch')
            for rowid in parent_rowids:
                yield np.empty(5 + rowid),

        @depc.register_preproc(
            'spam', ['fgweight', 'chip', 'keypoint'],
            ['spam', 'eggs', 'size', 'uuid', 'vector', 'textdata'],
            [str, int, (int, int), uuid.UUID, np.ndarray, ('extern', ut.readfrom)],
            docstr='I dont like spam',)
        def dummy_preproc_spam(depc, *args, **kwargs):
            config = kwargs.get('config', None)
            if config is None:
                config = {}
            print('[preproc] Computing spam')
            ut.writeto('tmp.txt', ut.lorium_ipsum())
            for x in zip(*args):
                size = (42, 21)
                uuid = ut.get_zero_uuid()
                vector = np.ones(3)
                yield ('spam', 3665, size, uuid, vector, 'tmp.txt')

        @depc.register_preproc(
            'nnindexer', ['keypoint*'], ['flann'], [str],  # [('extern', ut.load_data)],
            configclass=DummyIndexerConfig,
        )
        def dummy_preproc_indexer(depc, parent_rowids_list, config=None):
            print('COMPUTING DUMMY INDEXER')
            #assert len(parent_rowids_list) == 1, 'handles only one indexer'
            for parent_rowids in parent_rowids_list:
                yield ('really cool flann object' + str(config.get_cfgstr()) + ' ' + str(parent_rowids),)

        @depc.register_preproc(
            'notchpair', ['notch', 'notch'], ['pairscore'], [int],  # [('extern', ut.load_data)],
            #configclass=DummyIndexerConfig,
        )
        def dummy_notchpair(depc, n1, n2, config=None):
            print('COMPUTING MULTITEST 1 ')
            #assert len(parent_rowids_list) == 1, 'handles only one indexer'
            for nn1, nn2 in zip(n1, n2):
                yield (nn1 + nn2,)

        @depc.register_preproc(
            'multitest', ['keypoint', 'notch', 'notch', 'fgweight*', 'notchpair*', 'notchpair*', 'notchpair', 'nnindexer'], ['foo'], [str],  # [('extern', ut.load_data)],
            #configclass=DummyIndexerConfig,
        )
        def dummy_multitest(depc, *args, **kwargs):
            print('COMPUTING MULTITEST 1 ')
            #assert len(parent_rowids_list) == 1, 'handles only one indexer'
            for x in zip(args):
                yield ('cool multi object' + str(kwargs) + ' ' + str(x),)

        # TEST MULTISET DEPENDENCIES
        @depc.register_preproc(
            'multitest_score', ['multitest'], ['score'], [int],  # [('extern', ut.load_data)],
            #configclass=DummyIndexerConfig,
        )
        def dummy_multitest_score(depc, parent_rowids, config=None):
            print('COMPUTING DEPENDENCY OF MULTITEST 1 ')
            #assert len(parent_rowids_list) == 1, 'handles only one indexer'
            for parent_rowids in zip(parent_rowids):
                yield (parent_rowids,)

        # TEST MULTISET DEPENDENCIES
        @depc.register_preproc(
            'multitest_score_x', ['multitest_score', 'multitest_score'], ['score'], [int],  # [('extern', ut.load_data)],
            #configclass=DummyIndexerConfig,
        )
        def multitest_score_x(depc, *args, **kwargs):
            raise NotImplementedError('hack')
        # REGISTER MATCHING ALGORITHMS

        @depc.register_preproc(tablename='neighbs', colnames=['qx2_idx', 'qx2_dist'],
                               coltypes=[np.ndarray, np.ndarray],
                               parents=['keypoint', 'fgweight', 'nnindexer', 'nnindexer'])
        def neighbs(depc, *args, **kwargs):
            """
            CommandLine:
                python -m dtool_ibeis.base --exec-VsManySimilarityRequest
            """
            #dummy_preproc_kpts
            for qaid in zip(args):
                yield np.array([qaid]), np.array([qaid])

        @depc.register_preproc(tablename='neighbs_score', colnames=['qx2_dist'],
                               coltypes=[np.ndarray],
                               parents=['neighbs'])
        def neighbs_score(depc, *args, **kwargs):
            """
            CommandLine:
                python -m dtool_ibeis.base --exec-VsManySimilarityRequest
            """
            raise NotImplementedError('hack')

        @depc.register_preproc(
            'vsone', [dummy_root, dummy_root],
            ['score', 'match_obj', 'fm'],
            [float, DummyVsOneMatch, np.ndarray],
            requestclass=DummyVsOneRequest,
            configclass=DummyVsOneConfig,
            chunksize=2
        )
        def compute_vsone_matching(depc, qaids, daids, config):
            """
            CommandLine:
                python -m dtool_ibeis.base --exec-VsOneSimilarityRequest
            """
            print('RUNNING DUMMY VSONE ALGO')
            for qaid, daid in zip(qaids, daids):
                match = DummyVsOneMatch()
                match.qaid = qaid
                match.daid = daid
                match.fm = np.array([[1, 2], [3, 4]])
                score = match.score = qaid + daid
                yield (score, match, match.fm)

    # table = depc['spam']
    # print(ut.repr2(table.get_addtable_kw(), nl=2))
    depc.initialize()
    # table.print_schemadef()
    # print(table.db.get_schema_current_autogeneration_str())
    return depc


def example_getter_methods(depc, tablename, root_rowids):
    """
    example of different ways to get data
    """
    import dtool_ibeis
    print('\n+---')
    print('Running getter example')
    print(' * tablename=%r' % (tablename))
    print(' * root_rowids=%r' % (ut.trunc_repr(tablename)))

    # You can get a reference to data rows using the "root" (dummy_annot) rowids
    # By default, if the data has not been computed, then it will be computed
    # for you. But if you specify ensure=False, None will be returned if the data
    # has not been computed yet.
    tbl_rowids = depc.get_rowids(tablename, root_rowids, ensure=False)  # NOQA
    print('tbl_rowids = depc.get_rowids(tablename, root_rowids, ensure=False)')
    print('tbl_rowids = %s' % (ut.trunc_repr(tbl_rowids),))
    #assert tbl_rowids[0] is None

    # The default is for the data to be computed though. Manaual interactions will
    # launch as necessary.
    tbl_rowids = depc.get_rowids(tablename, root_rowids, ensure=True)  # NOQA
    print('tbl_rowids = depc.get_rowids(tablename, root_rowids, ensure=True)')
    print('tbl_rowids = %s' % (ut.trunc_repr(tbl_rowids),))
    assert tbl_rowids[0] is not None

    # Now the data is cached and will not need to be computed again
    tbl_rowids = depc.get_rowids(tablename, root_rowids, ensure=False)  # NOQA
    assert tbl_rowids[0] is not None

    # Can lookup a table, which can access data directly.  The rowids can be
    # used to lookup data values directly. By default all data in a row is
    # returned.
    table = depc[tablename]
    datas = table.get_row_data(tbl_rowids)  # NOQA

    # But you can also ask for a specific column
    col1 = table.columns[0]
    col1_data = table.get_row_data(tbl_rowids, col1)  # NOQA

    # In the case of external columns:
    if len(table.extern_columns) > 0:
        excol = table.extern_columns[0]
        # you can lookup the value of the external data very simply
        extern_data = table.get_row_data(tbl_rowids, (excol,))  # NOQA
        print('extern_data = table.get_row_data(tbl_rowids, (excol,))')
        print(ut.varinfo_str(extern_data, 'extern_data'))
        # you can lookup the hidden paths as follows
        extern_paths = table.get_row_data(tbl_rowids, (excol + dtool_ibeis.depcache_table.EXTERN_SUFFIX,))  # NOQA
        print('extern_paths = table.get_row_data(tbl_rowids, (excol + dtool_ibeis.depcache_table.EXTERN_SUFFIX,))')
        print(ut.varinfo_str(extern_paths, 'extern_paths'))

    # But you can also just the root rowids directly. This is the simplest way
    # to access data and really "all you need to know"
    if len(table.columns) > 1:
        col1, col2 = table.columns[0:2]
        datas = depc.get_property(tablename, root_rowids, (col1, col2))  # NOQA

    print('L__')


def test_getters(depc):
    # One input = one output
    chip = depc.get_property('chip', 1, 'chip')  # NOQA
    print('[test] chip.sum() = %r' % (chip.sum(),))

    col_tup_list = depc.get_property('chip', [1], ('size',))
    print('[test] col_tup_list = %r' % (col_tup_list,))

    col_list = depc.get_property('chip', [1], 'size')
    print('[test] col_list = %r' % (col_list,))

    col = depc.get_property('chip', 1, 'size')
    print('[test] col = %r' % (col,))

    cols = depc.get_property('chip', 1, 'size')
    print('[test] cols = %r' % (cols,))

    if False:
        chip_dict = depc.get_obj('chip', 1)

        print('chip_dict = %r' % (chip_dict,))
        for key in chip_dict.keys():
            print(ut.varinfo_str(chip_dict[key], 'chip_dict["%s"]' % (key,)))
        # print('chip_dict["chip"] = %s' % (ut.trunc_repr(chip_dict['chip']),))
        print('chip_dict = %r' % (chip_dict,))


def dummy_example_depcacahe():
    r"""
    CommandLine:
        python -m dtool_ibeis.example_depcache --exec-dummy_example_depcacahe

    Example:
        >>> # ENABLE_DOCTEST
        >>> from dtool_ibeis.example_depcache import *  # NOQA
        >>> depc = dummy_example_depcacahe()
    """
    fname = None
    # fname = 'dummy_default_depcache'
    fname = ':memory:'

    depc = testdata_depc(fname)

    tablename = 'fgweight'
    # print('[test] fgweight_path =\n%s' % (ut.repr3(depc.get_dependencies(tablename), nl=1),))
    # print('[test] keypoint =\n%s' % (ut.repr3(depc.get_dependencies('keypoint'), nl=1),))
    # print('[test] descriptor =\n%s' % (ut.repr3(depc.get_dependencies('descriptor'), nl=1),))
    # print('[test] spam =\n%s' % (ut.repr3(depc.get_dependencies('spam'), nl=1),))

    root_rowids = [5, 3]
    desc_rowids = depc.get_rowids('descriptor', root_rowids)  # NOQA

    table = depc[tablename]  # NOQA

    #example_getter_methods(depc, 'vsmany', root_rowids)
    # example_getter_methods(depc, 'chipmask', root_rowids)
    # example_getter_methods(depc, 'keypoint', root_rowids)
    # example_getter_methods(depc, 'chip', root_rowids)

    test_getters(depc)

    #import plottool_ibeis as pt
    # pt.ensureqt()

    graph = depc.make_graph()  # NOQA
    #pt.show_nx(graph)

    print('---------- 111 -----------')

    # Try testing the algorithm
    req = depc.new_request('vsmany', root_rowids, root_rowids, {})
    print('req = %r' % (req,))
    req.execute()

    print('---------- 222 -----------')

    cfgdict = {'sver_on': False}
    req = depc.new_request('vsmany', root_rowids, [root_rowids], cfgdict)
    req.execute()

    print('---------- 333 -----------')

    cfgdict = {'sver_on': False, 'adapt_shape': False}
    req = depc.new_request('vsmany', root_rowids, root_rowids, cfgdict)
    req.execute()

    print('---------- 444 -----------')

    req = depc.new_request('vsmany', root_rowids, root_rowids, {})
    req.execute()

    #ut.InstanceList(
    db = list(depc.fname_to_db.values())[0]
    #db_list = ut.InstanceList(depc.fname_to_db.values())
    #db_list.print_table_csv('config', exclude_columns='config_strid')

    print('config table')
    tablename = 'config'
    column_list, column_names = db.get_table_column_data(tablename,
                                                         ['config_strid'])
    print('\n'.join([ut.hz_str(*list(ut.interleave((r, [', '] * (len(r) - 1)))))
                     for r in list(zip(*[[ut.repr3(r, nl=2) for r in col] for col in column_list]))]))

    return depc

if __name__ == '__main__':
    r"""
    CommandLine:
        python -m dtool_ibeis.example_depcache
        python -m dtool_ibeis.example_depcache --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# from __future__ import absolute_import, division, print_function, unicode_literals
# import utool as ut
# import numpy as np
# from six.moves import zip
# from dtool_ibeis import depcache_table


# def get_all_descendant_rowids(depc, tablename, root_rowids, config=None,
#                               ensure=True, eager=True, nInput=None,
#                               recompute=False, recompute_all=False,
#                               levels_up=None, _debug=False):
#     r"""
#     Connects `root_rowids` to rowids in `tablename`, and computes all
#     values needed along the way. This is the main workhorse function for
#     dependency computations.

#     Args:
#         tablename (str): table to compute dependencies to
#         root_rowids (list): rowids for ``tablename``
#         config (dict): config applicable for all tables (default = None)
#         ensure (bool): eager evaluation if True(default = True)
#         eager (bool): (default = True)
#         nInput (None): (default = None)
#         recompute (bool): (default = False)
#         recompute_all (bool): (default = False)
#         levels_up (int): only partially compute dependencies (default = 0)
#         _debug (bool): (default = False)

#     CommandLine:
#         python -m dtool_ibeis.depcache_control --exec-get_all_descendant_rowids:0
#         python -m dtool_ibeis.depcache_control --exec-get_all_descendant_rowids:1

#     Example:
#         >>> # DISABLE_DOCTEST
#         >>> from dtool_ibeis.depcache_control import *  # NOQA
#         >>> from dtool_ibeis.example_depcache import testdata_depc
#         >>> depc = testdata_depc()
#         >>> tablename = 'spam'
#         >>> root_rowids = [1, 2]
#         >>> config1 = {'dim_size': 500}
#         >>> config2 = {'dim_size': 100}
#         >>> config3 = {'dim_size': 500, 'adapt_shape': False}
#         >>> ensure, eager, nInput = True, True, None
#         >>> _debug = True
#         >>> rowid_dict1 = depc.get_all_descendant_rowids(
#         >>>     tablename, root_rowids, config1, ensure, eager, nInput, _debug=_debug)
#         >>> rowid_dict2 = depc.get_all_descendant_rowids(
#         >>>     tablename, root_rowids, config2, ensure, eager, nInput, _debug=_debug)
#         >>> rowid_dict3 = depc.get_all_descendant_rowids(
#         >>>     tablename, root_rowids, config3, ensure, eager, nInput, _debug=_debug)
#         >>> result1 = 'rowid_dict1 = ' + ut.repr3(rowid_dict1, nl=1)
#         >>> result2 = 'rowid_dict2 = ' + ut.repr3(rowid_dict2, nl=1)
#         >>> result3 = 'rowid_dict3 = ' + ut.repr3(rowid_dict3, nl=1)
#         >>> result = '\n'.join([result1, result2, result3])
#         >>> print(result)
#         rowid_dict1 = {
#             'chip': [1, 2],
#             'dummy_annot': [1, 2],
#             'fgweight': [1, 2],
#             'keypoint': [1, 2],
#             'probchip': [1, 2],
#             'spam': [1, 2],
#         }
#         rowid_dict2 = {
#             'chip': [3, 4],
#             'dummy_annot': [1, 2],
#             'fgweight': [3, 4],
#             'keypoint': [3, 4],
#             'probchip': [1, 2],
#             'spam': [3, 4],
#         }
#         rowid_dict3 = {
#             'chip': [1, 2],
#             'dummy_annot': [1, 2],
#             'fgweight': [5, 6],
#             'keypoint': [5, 6],
#             'probchip': [1, 2],
#             'spam': [5, 6],
#         }


#     Example:
#         >>> # ENABLE_DOCTEST
#         >>> from dtool_ibeis.depcache_control import *  # NOQA
#         >>> from dtool_ibeis.example_depcache import testdata_depc
#         >>> depc = testdata_depc()
#         >>> _debug = True
#         >>> tablename = 'vsmany'
#         >>> config = depc.configclass_dict['vsmany']()
#         >>> root_rowids = [1, 2, 3]
#         >>> ensure, eager, nInput = False, True, None
#         >>> # Get rowids of algo ( should be None )
#         >>> rowid_dict = depc.get_all_descendant_rowids(
#         >>>     tablename, root_rowids, config, ensure, eager, nInput,
#         >>>     _debug=_debug)
#         >>> result = ut.repr3(rowid_dict, nl=1)
#         >>> print(result)
#         {
#             'dummy_annot': [1, 2, 3],
#             'vsmany': [None, None, None],
#         }

#     Example:
#         >>> # ENABLE_DOCTEST
#         >>> from dtool_ibeis.depcache_control import *  # NOQA
#         >>> from dtool_ibeis.example_depcache import testdata_depc
#         >>> # Make sure algo config can correctly get properites
#         >>> depc = testdata_depc()
#         >>> tablename = 'chip'
#         >>> recompute = False
#         >>> recompute_all = False
#         >>> _debug = True
#         >>> root_rowids = [1, 2]
#         >>> configclass = depc.configclass_dict['chip']
#         >>> config_ = configclass()
#         >>> config1 = depc.configclass_dict['vsmany'](dim_size=500)
#         >>> config2 = depc.configclass_dict['vsmany'](dim_size=100)
#         >>> config = config2
#         >>> prop_dicts1 = depc.get_all_descendant_rowids(
#         >>>     tablename, root_rowids, config=config1, _debug=_debug)
#         >>> prop_dicts2 = depc.get_all_descendant_rowids(
#         >>>     tablename, root_rowids, config=config2, _debug=_debug)
#         >>> print(prop_dicts2)
#         >>> print(prop_dicts1)
#         >>> assert prop_dicts1 != prop_dicts2

#     Example:
#         >>> # ENABLE_DOCTEST
#         >>> from dtool_ibeis.depcache_control import *  # NOQA
#         >>> from dtool_ibeis.example_depcache import testdata_depc
#         >>> depc = testdata_depc()
#         >>> exec(ut.execstr_funckw(depc.get_all_descendant_rowids), globals())
#         >>> _debug = True
#         >>> qaids, daids = [1, 2, 4], [2, 3, 4]
#         >>> root_rowids = list(zip(*ut.product(qaids, daids)))
#         >>> request = depc.new_request('vsone', qaids, daids)
#         >>> results = request.execute()
#         >>> tablename = 'vsone'
#         >>> rowid_dict = depc.get_all_descendant_rowids(
#         >>>     tablename, root_rowids, config=None, _debug=_debug)
#     """
#     # TODO: Need to have a nice way of ensuring configs dont overlap
#     # via namespaces.
#     _debug = depc._debug if _debug is None else _debug
#     indenter = ut.Indenter('[Descend-to-%s]' % (tablename,), enabled=_debug)
#     if _debug:
#         indenter.start()
#         print(' * GET DESCENDANT ROWIDS %s ' % (tablename,))
#         print(' * config = %r' % (config,))
#     dependency_levels = depc.get_dependencies(tablename)
#     if levels_up is not None:
#         dependency_levels = dependency_levels[:-levels_up]

#     configclass_levels = [
#         [depc.configclass_dict.get(tablekey, None)
#          for tablekey in keys]
#         for keys in dependency_levels
#     ]
#     if _debug:
#         print('[depc] dependency_levels = %s' %
#               ut.repr3(dependency_levels, nl=1))
#         print('[depc] config_levels = %s' %
#               ut.repr3(configclass_levels, nl=1))

#     # TODO: better support for multi-edges
#     if (len(root_rowids) > 0 and ut.isiterable(root_rowids[0]) and
#          not depc[tablename].ismulti):
#         rowid_dict = {}
#         for colx, col in enumerate(root_rowids):
#             rowid_dict[depc.root + '%d' % (colx + 1,)] = col
#         rowid_dict[depc.root] = ut.unique_ordered(ut.flatten(root_rowids))
#     else:
#         rowid_dict = {depc.root: root_rowids}

#     # Ensure that each level ``tablename``'s dependencies have been computed
#     for level_keys in dependency_levels[1:]:
#         if _debug:
#             print(' * level_keys %s ' % (level_keys,))
#         # For each table in the level
#         for tablekey in level_keys:
#             try:
#                 child_rowids = depc._expand_level_rowids(
#                     tablename, tablekey, rowid_dict, ensure, eager, nInput,
#                     config, recompute, recompute_all, _debug)
#             except Exception as ex:
#                 table = depc[tablekey]  # NOQA
#                 keys = ['tablename', 'tablekey', 'rowid_dict', 'config',
#                         'table', 'dependency_levels']
#                 ut.printex(ex, 'error expanding rowids', keys=keys)
#                 raise
#             rowid_dict[tablekey] = child_rowids
#     if _debug:
#         print(' GOT DESCENDANT ROWIDS')
#         indenter.stop()
#     return rowid_dict


# def get_rowids(depc, tablename, root_rowids, config=None, ensure=True,
#                eager=True, nInput=None, _debug=None, recompute=False,
#                recompute_all=False):
#     """
#     Returns the rowids of `tablename` that correspond to `root_rowids`
#     using `config`.

#     Ignore:
#         tablename = 'nnindexer'
#         multi_rowids = (1, 2, 3, 4, 5)
#         root_rowids = [[multi_rowids]]
#         import plottool_ibeis as pt
#         pt.ensureqt()

#         from dtool_ibeis.depcache_control import *  # NOQA
#         from dtool_ibeis.example_depcache import testdata_depc
#         depc = testdata_depc()
#         exec(ut.execstr_funckw(depc.get_rowids), globals())
#         print(ut.depth_profile(root_rowids))
#         tablename = 'neighbs'
#         table = depc[tablename]  # NOQA
#         import plottool_ibeis as pt
#         pt.ensureqt()
#         _debug = depc._debug = True
#         depc.get_rowids(tablename, root_rowids, config, _debug=_debug)

#         pt.show_nx(depc.graph)
#         for key, val in table.type_to_subgraph.items():
#             pt.show_nx(val)
#             pt.set_title(key)

#     CommandLine:
#         python -m dtool_ibeis.depcache_control --exec-get_rowids
#         python -m dtool_ibeis.depcache_control --dump-get_rowids
#         python -m dtool_ibeis.depcache_control --exec-get_rowids:0

#     GridParams:
#         >>> param_grid = dict(
#         >>>     tablename=[ 'spam', 'neighbs'] # 'spam', 'multitest_score','keypoint'],
#         >>>   #tablename=['neighbs', 'keypoint', 'spam', 'multitest_score','keypoint'],
#         >>> )
#         >>> flat_root_ids = [1, 2, 3]
#         >>> combos = ut.all_dict_combinations(param_grid)
#         >>> index = 0
#         >>> keys = 'tablename'.split(', ')
#         >>> tablename, = ut.dict_take(combos[index], keys)

#     Setup:
#         >>> # DISABLE_GRID_DOCTEST
#         >>> from dtool_ibeis.depcache_control import *  # NOQA
#         >>> from dtool_ibeis.example_depcache import testdata_depc
#         >>> depc = testdata_depc()
#         >>> exec(ut.execstr_funckw(depc.get_rowids), globals())
#         >>> import plottool_ibeis as pt
#         >>> pt.ensureqt()
#         >>> #pt.show_nx(depc.graph)

#     GridExample0:
#         >>> table = depc[tablename]  # NOQA
#         >>> flat_root_ids = [1, 2, 3]
#         >>> root_rowids = [flat_root_ids for _ in table.input_order]
#         >>> print('root_rowids = %r' % (root_rowids,))
#         >>> #root_rowids = [[flat_root_ids], [(flat_root_ids,)]]
#         >>> #root_rowids = [list(zip(flat_root_ids)), (flat_root_ids,)]
#         >>> _debug = True
#         >>> depc.get_rowids(tablename, root_rowids, config, _debug=_debug)
#         >>> for key, val in table.type_to_subgraph.items():
#         >>>     pt.show_nx(val)
#         >>>     pt.set_title(key)

#     Example1:
#         >>> # ENABLE_DOCTEST
#         >>> from dtool_ibeis.depcache_control import *  # NOQA
#         >>> from dtool_ibeis.example_depcache import testdata_depc
#         >>> depc = testdata_depc()
#         >>> exec(ut.execstr_funckw(depc.get_rowids), globals())
#         >>> root_rowids = [1, 2, 3]
#         >>> tablename = 'spam'
#         >>> table = depc[tablename]
#         >>> kp_rowids = depc.get_rowids(tablename, root_rowids)
#         >>> #result = ('prop_list = %s' % (ut.repr2(prop_list),))
#         >>> #print(result)

#     Example:
#         >>> # ENABLE_DOCTEST
#         >>> from dtool_ibeis.depcache_control import *  # NOQA
#         >>> from dtool_ibeis.example_depcache import testdata_depc
#         >>> depc = testdata_depc()
#         >>> exec(ut.execstr_funckw(depc.get_rowids), globals())
#         >>> flat_root_ids = [1, 2, 3]
#         >>> kp_rowids = depc.get_rowids('keypoint', flat_root_ids)
#         >>> root_rowids = [flat_root_ids] * 8
#         >>> _debug = True
#         >>> tablename = 'nnindexer'
#         >>> tablename = 'multitest_score'
#         >>> table = depc[tablename]  # NOQA
#         >>> #result = ('prop_list = %s' % (ut.repr2(prop_list),))
#         >>> # print(result)
#     """
#     _debug = depc._debug if _debug is None else _debug
#     if _debug:
#         print(' * root_rowids=%s' % (ut.trunc_repr(root_rowids),))
#         print(' * config = %r' % (config,))
#     table = depc[tablename]  # NOQA
#     INDEXER_VERSION = False

#     if tablename == 'neighbor_index':
#         """
#         python -m ibeis.core_annots --exec-compute_neighbor_index --show
#         """

#         import utool
#         utool.embed()

#     if INDEXER_VERSION or tablename == 'neighbs':
#         compute_order = table.compute_order
#         depend_order = compute_order['depend_compute_ids']
#         input_order = compute_order['input_compute_ids']

#         if _debug:
#             print(' * input_order = %s' % (ut.repr3(input_order, nl=1),))
#             print(' * depend_order = %s' % (ut.repr3(depend_order, nl=1),))
#         if len(input_order) > 1:
#             assert ut.depth_atleast(root_rowids, 2), (
#                 'input_order = %r' % (input_order,))

#         with ut.Indenter('[GetRowID-%s]' % (tablename,),
#                          enabled=_debug):
#             # New way to get rowids
#             input_level = depend_order[0]
#             mid_levels = depend_order[1:-1]
#             output_level = depend_order[-1]

#             # List that holds a mapping from input order to input "name"
#             input_order_lookup = ut.make_index_lookup(input_order)
#             # Dictionary that holds the rowids computed for each table
#             # while tracing the dependencies.
#             rowid_lookup = ut.odict([(key, ut.odict()) for key in input_order])

#             # Need to split each path into parts.
#             # Each part represents another level of unflattening
#             # (because root indicies are all flat)

#             # Handle input level
#             assert input_level[0] == depc.root
#             for compute_id in input_order:
#                 # for name in input_names:
#                 argx = input_order_lookup[compute_id]
#                 rowid_lookup[compute_id] = root_rowids[argx]
#                 # HACK: Flatten to scalars
#                 # The inputs should just be given in the "correct" nesting.
#                 # TODO: determine what correct nesting is.
#                 for i in range(5):
#                     try:
#                         current = rowid_lookup[compute_id]
#                         rowid_lookup[compute_id] = ut.flatten(current)
#                     except Exception:
#                         pass

#             level = 0
#             if _debug:
#                 print('input_order_lookup = %r' % (input_order_lookup,))
#                 ut.printdict(rowid_lookup, 'rowid_lookup')

#             def handle_level(compute_id, rowid_lookup, _recompute, level):
#                 print('+--- HANDLE LEVEL %d -------' % (level,))
#                 tablekey = compute_id[0]
#                 input_suff = compute_id[1]
#                 config_ = depc._ensure_config(tablekey, config)
#                 table = depc[tablekey]
#                 lookupkeys = [(n, input_suff) for n in table.parent_id_tablenames]
#                 # ordering = ut.dict_take(input_order_lookup, input_names)
#                 # sortx = ut.argsort(ordering)
#                 # FIXME: get inputs for each table.
#                 # input_names = ut.take(input_names, sortx)
#                 # lookupkeys = list(ut.iprod(table.parent_id_tablenames, input_names))
#                 # lookupkeys = list(zip(table.parent_id_tablenames, input_types))
#                 if _debug:
#                     print('---- LOCALS ------')
#                     ut.print_locals(compute_id, tablekey, lookupkeys, table)
#                     print('L----------')
#                 # FIXME generalize
#                 _parent_ids = [rowid_lookup[tblkey] for tblkey in lookupkeys]
#                 if table.ismulti:
#                     parent_rowidsT = [[tuple(x)] for x in _parent_ids]
#                 else:
#                     parent_rowidsT = _parent_ids
#                 parent_rowidsT = np.broadcast_arrays(*parent_rowidsT)
#                 parent_rowids = list(zip(*parent_rowidsT))
#                 # Probably not right for general multi-input
#                 import utool
#                 with utool.embed_on_exception_context:
#                     next_rowids = table.get_rowid(
#                         parent_rowids, config=config_, eager=eager, nInput=nInput,
#                         ensure=ensure, recompute=_recompute)
#                 rowid_lookup[compute_id] = next_rowids
#                 if _debug:
#                     ut.printdict(rowid_lookup, 'rowid_lookup')
#                 if _debug:
#                     print('L___ HANDLE LEVEL %d -------' % (level,))
#                 return next_rowids

#             # Handle mid levels
#             _recompute = recompute_all
#             for level, compute_id in enumerate(mid_levels, start=1):
#                 handle_level(compute_id, rowid_lookup, _recompute, level)
#             level += 1

#             # Handel final (requested) level
#             compute_id = output_level
#             _recompute = recompute
#             rowid_list =  handle_level(compute_id, rowid_lookup,
#                                        _recompute, level)
#     else:
#         with ut.Indenter('[GetRowID-%s]' % (tablename,),
#                          enabled=_debug):
#             # TODO: Get nonself rowids first
#             # THen get self rowids for debugging ease
#             try:
#                 if False:
#                     recompute_ = recompute or recompute_all
#                     parent_rowids = depc._get_parent_input(
#                         tablename, root_rowids, config, ensure=True, _debug=None,
#                         recompute=False, recompute_all=False, eager=True,
#                         nInput=None)
#                     config_ = depc._ensure_config(tablename, config)
#                     #if onthefly:
#                     #    pass
#                     table = depc[tablename]
#                     rowid_list = table.get_rowid(
#                         parent_rowids, config=config_, eager=eager, nInput=nInput,
#                         ensure=ensure, recompute=recompute_)
#                 else:
#                     # Compute everything from the root to the requested table
#                     rowid_dict = depc.get_all_descendant_rowids(
#                         tablename, root_rowids, config=config, ensure=ensure,
#                         eager=eager, nInput=nInput, recompute=recompute,
#                         recompute_all=recompute_all, _debug=ut.countdown_flag(_debug))
#                     rowid_list = rowid_dict[tablename]
#             except depcache_table.ExternalStorageException:
#                 print('EXTERNAL EXCEPTION One retry in get_rowids')
#                 rowid_dict = depc.get_all_descendant_rowids(
#                     tablename, root_rowids, config=config, ensure=ensure,
#                     eager=eager, nInput=nInput, recompute=recompute,
#                     recompute_all=recompute_all, _debug=ut.countdown_flag(_debug))
#                 rowid_list = rowid_dict[tablename]
#     if _debug:
#         print(' * return rowid_list = %s' % (ut.trunc_repr(rowid_list),))
#     return rowid_list


# def _get_parent_input(depc, tablename, root_rowids, config, ensure=True,
#                       _debug=None, recompute=False, recompute_all=False,
#                       eager=True, nInput=None):
#     # Get ancestor rowids that are descendants of root
#     table = depc[tablename]
#     rowid_dict = depc.get_all_descendant_rowids(
#         tablename, root_rowids, config=config, ensure=ensure,
#         eager=eager, nInput=nInput, recompute=recompute,
#         recompute_all=recompute_all, _debug=ut.countdown_flag(_debug),
#         levels_up=1)
#     parent_rowids = depc._get_parent_rowids(table, rowid_dict)
#     return parent_rowids

# #def get_relevant_subconfigs(depc, tablename, config):
# #    depc._ensure_config(tablename, config)
# #    pass


# def _get_parent_rowids(depc, table, rowid_dict):
#     # FIXME to handle multiedges correctly
#     parent_rowidsT = ut.dict_take(rowid_dict,
#                                   table.parent_id_tablenames)
#     if table.ismulti:
#         parent_rowids = parent_rowidsT
#     else:
#         parent_rowids = ut.list_transpose(parent_rowidsT)
#     return parent_rowids


# def _expand_level_rowids(depc, tablename, tablekey, rowid_dict, ensure,
#                          eager, nInput, config, recompute, recompute_all,
#                          _debug):
#     table = depc[tablekey]
#     config_ = depc._ensure_config(tablekey, config)
#     parent_rowids = depc._get_parent_rowids(table, rowid_dict)
#     if _debug:
#         print('   * tablekey = %r' % (tablekey,))
#         print('   * (ensured) config_ = %r' % (config_,))
#         print('   * config_rowid = %r' % (table.get_config_rowid(config_),))
#         print('   * parent_rowids = %s' % (ut.trunc_repr(parent_rowids),))
#     _recompute = recompute_all or (tablekey == tablename and recompute)
#     level_rowids = table.get_rowid(
#         parent_rowids, config=config_, eager=eager, nInput=nInput,
#         ensure=ensure, recompute=_recompute)
#     if _debug:
#         print('   * level_rowids = %s' % (ut.trunc_repr(level_rowids),))
#     r


# def test_getprop_with_configs():
#     r"""
#     CommandLine:
#         python -m dtool_ibeis.example_depcache2 test_getprop_with_configs --show

#     Example:
#         >>> # DISABLE_DOCTEST
#         >>> from dtool_ibeis.example_depcache2 import *  # NOQA
#         >>> test_getprop_with_configs()
#     """
#     config1 = {'manual_extract': True}
#     config2 = {'manual_extract': False}
#     depc = testdata_depc2()

#     aid = 2
#     _debug = False

#     cropchip1 = depc.get('cropchip', aid, 'img', config=config1)
#     cropchip2 = depc.get('cropchip', aid, 'img', config=config2)
#     print('cropchip1.shape = %r' % (cropchip1.shape,))
#     print('cropchip2.shape = %r' % (cropchip2.shape,))
#     cropchip1 = depc.get('cropchip', aid, 'img', config=config1)
#     cropchip2 = depc.get('cropchip', aid, 'img', config=config2)

#     print('cropchip1.shape = %r' % (cropchip1.shape,))
#     print('cropchip2.shape = %r' % (cropchip2.shape,))

#     chip = depc.get('chip', aid, 'img')
#     print('chip.shape = %r' % (chip.shape,))

#     tip1 = depc.get('tip', aid, config=config1, _debug=_debug)
#     tip2 = depc.get('tip', aid, config=config2, _debug=_debug)

#     print('tip1 = %r' % (tip1,))
#     print('tip2 = %r' % (tip2,))

#     depc.print_all_tables()
#     depc.print_config_tables()
#     #import utool
#     #utool.embed()


# def testdata_depc2():
#     """
#     Example of local registration
#     sudo pip install freetype-py

#     CommandLine:
#         python -m dtool_ibeis.example_depcache2 testdata_depc2 --show

#     Example:
#         >>> # DISABLE_DOCTEST
#         >>> from dtool_ibeis.example_depcache2 import *  # NOQA
#         >>> depc = testdata_depc2()
#         >>> ut.quit_if_noshow()
#         >>> import plottool_ibeis as pt
#         >>> depc.show_graph()
#         >>> ut.show_if_requested()
#     """
#     import dtool_ibeis
#     import vtool_ibeis as vt
#     from vtool_ibeis import fontdemo

#     # put the test cache in the dtool_ibeis repo
#     dtool_repo = dirname(ut.get_module_dir(dtool_ibeis))
#     cache_dpath = join(dtool_repo, 'DEPCACHE2')

#     root = 'annot'

#     depc = dtool_ibeis.DependencyCache(
#         root_tablename=root, cache_dpath=cache_dpath, use_globals=False)

#     # ----------

#     class ChipConfig(dtool_ibeis.Config):
#         _param_info_list = [
#             ut.ParamInfo('dim_size', 500),
#             ut.ParamInfo('ext', '.png'),
#         ]

#     @depc.register_preproc(
#         tablename='chip', parents=[root], colnames=['size', 'img'],
#         coltypes=[(int, int), ('extern', vt.imread, vt.imwrite)],
#         configclass=ChipConfig)
#     def compute_chip(depc, aids, config=None):
#         for aid in aids:
#             chip = fontdemo.get_text_test_img(str(aid))
#             size = vt.get_size(chip)
#             yield size, chip

#     # ----------

#     class TipConfig(dtool_ibeis.Config):
#         _param_info_list = [
#             ut.ParamInfo('manual_extract', False, hideif=False),
#         ]

#     @depc.register_preproc(
#         tablename='tip', parents=['chip'],
#         colnames=['notch', 'left', 'right'],
#         coltypes=[np.ndarray, np.ndarray, np.ndarray],
#         configclass=TipConfig,
#     )
#     def compute_tips(depc, chip_rowids, config=None):
#         manual_extract = config['manual_extract']
#         chips = depc.get_native('chip', chip_rowids, 'img')
#         for chip in chips:
#             seed = (chip).sum()
#             perb = ((seed % 1000) / 1000) * .25
#             w, h = vt.get_size(chip)
#             if manual_extract:
#                 # Make noticable difference between config outputs
#                 lpb =  np.ceil(w * perb)
#                 npb =  np.ceil(h * perb)
#                 rpb = -np.ceil(w * perb)
#             else:
#                 lpb =  np.ceil(w * perb / 2)
#                 npb = -np.ceil(h * perb)
#                 rpb = -np.ceil(w * perb)
#             wh = np.array([w, h], dtype=np.int32)[None, :]
#             rel_base = np.array([[.0, .5], [.5, .5], [1., .5]])
#             offset   = np.array([[lpb, 0], [0, npb], [rpb, 0]])
#             tip = np.round((wh * rel_base)) + offset
#             left, notch, right = tip
#             yield left, notch, right

#     # ----------

#     class CropChipConfig(dtool_ibeis.Config):
#         _param_info_list = [
#             ut.ParamInfo('dim_size', 500),
#         ]

#     @depc.register_preproc(
#         tablename='cropchip', parents=['chip', 'tip'],
#         colnames=['img'],
#         coltypes=[np.ndarray],
#         configclass=CropChipConfig,
#     )
#     def compute_cropchip(depc, cids, tids, config=None):
#         print("COMPUTE CROPCHIP")
#         print('config = %r' % (config,))
#         chips = depc.get_native('chip', cids, 'img')
#         tips = depc.get_native('tip', tids)
#         print('tips = %r' % (tips,))
#         for chip, tip in zip(chips, tips):
#             notch, left, right = tip
#             lx = left[0]
#             rx = right[0]
#             cropped_chip = chip[lx:(rx - 1), ...]
#             yield (cropped_chip,)

#     # ----------

#     class TrailingEdgeConfig(dtool_ibeis.Config):
#         _param_info_list = []

#     @depc.register_preproc(
#         tablename='trailingedge', parents=['cropchip'],
#         colnames=['te'],
#         coltypes=[np.ndarray],
#         configclass=TrailingEdgeConfig,
#     )
#     def compute_trailing_edge(depc, cropped_chips, config=None):
#         for cc in cropped_chips:
#             #depc.get_native('chip', cids)
#             size = 1
#             te = np.arange(size)
#             yield (te,)

#     depc.initialize()
#     return depc


# def testdata_depc_image():
#     """
#     Example of local registration
#     sudo pip install freetype-py

#     CommandLine:
#         python -m dtool_ibeis.example_depcache2 testdata_depc_image --show

#     Example:
#         >>> # DISABLE_DOCTEST
#         >>> from dtool_ibeis.example_depcache2 import *  # NOQA
#         >>> depc = testdata_depc_image()
#         >>> ut.quit_if_noshow()
#         >>> import plottool_ibeis as pt
#         >>> depc.show_graph()
#         >>> depc['detection'].show_input_graph()
#         >>> print(depc['detection'].compute_order)
#         >>> ut.show_if_requested()
#     """
#     import dtool_ibeis

#     # put the test cache in the dtool_ibeis repo
#     dtool_repo = dirname(ut.get_module_dir(dtool_ibeis))
#     cache_dpath = join(dtool_repo, 'DEPCACHE2')

#     root = 'image'

#     depc = dtool_ibeis.DependencyCache(
#         root_tablename=root, cache_dpath=cache_dpath, use_globals=False)

#     # ----------
#     dummy_cols = dict(colnames=['data'], coltypes=[np.ndarray])
#     def dummy_func(depc, *args, **kwargs):
#         for row_arg in zip(*args):
#             yield (np.array([42]),)

#     depc.register_preproc(tablename='detector', parents=['image*'], **dummy_cols)(dummy_func)
#     depc.register_preproc(tablename='detection', parents=['image', 'detector'], **dummy_cols)(dummy_func)

#     depc.initialize()
#     return depc


# def testdata_depc_annot():
#     """
#     CommandLine:
#         python -m dtool_ibeis.example_depcache2 testdata_depc_annot --show

#     Example:
#         >>> # DISABLE_DOCTEST
#         >>> from dtool_ibeis.example_depcache2 import *  # NOQA
#         >>> depc = testdata_depc_annot()
#         >>> ut.quit_if_noshow()
#         >>> import plottool_ibeis as pt
#         >>> depc.show_graph()
#         >>> tablename = 'featweight'
#         >>> table = depc[tablename]
#         >>> table.show_input_graph()
#         >>> print(table.compute_order)
#         >>> ut.show_if_requested()
#     """
#     import dtool_ibeis
#     # put the test cache in the dtool_ibeis repo
#     dtool_repo = dirname(ut.get_module_dir(dtool_ibeis))
#     cache_dpath = join(dtool_repo, 'DEPCACHE2')
#     dummy_cols = dict(colnames=['data'], coltypes=[np.ndarray])
#     def dummy_func(depc, *args, **kwargs):
#         for row_arg in zip(*args):
#             yield (np.array([42]),)

#     # NOTE: Consider the smk_match.
#     # It would be really cool if we could say that the vocab
#     # for the input to the parent smk_vec must be the same vocab
#     # that was used to compute the inverted index. How do we encode that?

#     root = 'annot'
#     #vocab_parent = 'annot'
#     #vocab_parent = 'chip'
#     #vocab_parent = 'feat'
#     vocab_parent = 'featweight'
#     depc = dtool_ibeis.DependencyCache(
#         root_tablename=root, cache_dpath=cache_dpath, use_globals=False)
#     depc.register_preproc(tablename='chip', parents=['annot'], **dummy_cols)(dummy_func)
#     depc.register_preproc(tablename='fgmodel', parents=['chip*'], **dummy_cols)(dummy_func)
#     depc.register_preproc(tablename='probchip', parents=['annot', 'fgmodel'], **dummy_cols)(dummy_func)
#     #depc.register_preproc(tablename='probchip', parents=['annot'], **dummy_cols)(dummy_func)
#     depc.register_preproc(tablename='feat', parents=['chip'], **dummy_cols)(dummy_func)
#     #depc.register_preproc(tablename='feat', parents=['annot'], **dummy_cols)(dummy_func)
#     depc.register_preproc(tablename='featweight', parents=['feat', 'probchip'], **dummy_cols)(dummy_func)

#     depc.register_preproc(tablename='indexer', parents=[vocab_parent + '*'], **dummy_cols)(dummy_func)
#     depc.register_preproc(tablename='neighbs', parents=[vocab_parent, 'indexer'], **dummy_cols)(dummy_func)
#     depc.register_preproc(tablename='vocab', parents=[vocab_parent + '*'], **dummy_cols)(dummy_func)
#     depc.register_preproc(tablename='smk_vec', parents=[vocab_parent, 'vocab'], **dummy_cols)(dummy_func)
#     depc.register_preproc(tablename='inv_index', parents=['smk_vec*'], **dummy_cols)(dummy_func)
#     depc.register_preproc(tablename='smk_match', parents=['smk_vec', 'inv_index'], **dummy_cols)(dummy_func)
#     depc.register_preproc(tablename='vsone', parents=[vocab_parent, vocab_parent], **dummy_cols)(dummy_func)

#     depc.register_preproc(tablename='viewpoint_model', parents=['annot*'], **dummy_cols)(dummy_func)
#     depc.register_preproc(tablename='viewpoint', parents=['annot', 'viewpoint_model'], **dummy_cols)(dummy_func)

#     depc.register_preproc(tablename='quality_model', parents=['annot*'], **dummy_cols)(dummy_func)
#     depc.register_preproc(tablename='quality', parents=['annot', 'quality_model'], **dummy_cols)(dummy_func)

#     depc.initialize()
#     return depc

# return level_rowids



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Initially Generated By:
    python -m utool --tf setup_repo --repo=dtool_ibeis --codedir=~/code --modname=dtool_ibeis
"""
from __future__ import absolute_import, division, print_function, unicode_literals


def dtool_main():
    ignore_prefix = []
    ignore_suffix = []
    import utool as ut
    # allows for --tf
    ut.main_function_tester('dtool_ibeis', ignore_prefix, ignore_suffix)

if __name__ == '__main__':
    """
    Usage:
        python -m dtool_ibeis --tf <funcname>
    """
    print('Running dtool_ibeis main')
    dtool_main()




# @ut.accepts_scalar_input2(argx_list=[1])
# def get_obj(depc, tablename, root_rowids, config=None, ensure=True):
#     """ Convinience function. Gets data in `tablename` as a list of
#     objects. """
#     print('WARNING EXPERIMENTAL')
#     try:
#         if tablename == depc.root:
#             obj_list = list(depc._root_asobject(root_rowids))
#         else:
#             def make_property_getter(rowid, colname):
#                 def wrapper():
#                     return depc.get_property(
#                         tablename, rowid, colnames=colname, config=config,
#                         ensure=ensure)
#                 return wrapper
#             colnames = depc[tablename].data_colnames
#             obj_list = [
#                 ut.LazyDict({colname: make_property_getter(rowid, colname)
#                              for colname in colnames})
#                 for rowid in root_rowids
#             ]
#         return obj_list
#         # data_list = depc.get_property(tablename, root_rowids, config)
#         # # TODO: lazy dict
#         # return [dict(zip(colnames, data)) for data in data_list]
#     except Exception as ex:
#         ut.printex(ex, 'error in getobj', keys=['tablename', 'root_rowids',
#                                                 'colnames'])
#         raise


#     def root_asobject(aid_list):
#         """ Convinience for writing preproc funcs """
#         for aid in aid_list:
#             gpath = gpath_list[aid]
#             root_obj = ut.LazyDict({
#                 'aid': aid,
#                 'gpath': gpath,
#                 'image': lambda: vt.imread(gpath)
#             })
#             yield root_obj


#     @depc.register_preproc(
#         tablename='chip', parents=[dummy_root], colnames=['size', 'chip'],
#         coltypes=[(int, int), vt.imread], configclass=DummyChipConfig,
#         asobject=True)
#     def dummy_preproc_chip(depc, annot_list, config=None):
#         """
#         TODO: Infer properties from docstr

#         Args:
#             annot_list (list): list of annot objects
#             config (dict): config dictionary

#         Returns:
#             tuple : ((int, int), ('extern', vt.imread))
#         """
#         if config is None:
#             config = {}
#         # Demonstates using asobject to get input to function as a dictionary
#         # of properties
#         for annot in annot_list:
#             print('[preproc] Computing chips of annot=%r' % (annot,))
#             chip_fpath = annot['gpath']
#             w, h = vt.image.open_image_size(chip_fpath)
#             size = (w, h)
#             print('* chip_fpath = %r' % (chip_fpath,))
#             print('* size = %r' % (size,))
#             yield size, chip_fpath

# #config_hashid = config.get('feat_cfgstr')
# #assert config_hashid is not None
# # TODO store config_rowid in qparams
# #else:
# #    config_hashid = db.cfg.feat_cfg.get_cfgstr()
# if False:
#     if config is not None:
#         try:
#             #config_hashid = 'none'
#             config_hashid = config.get(table.tablename + '_hashid')
#         except KeyError:
#             try:
#                 subconfig = config.get(table.tablename + '_config')
#                 config_hashid = ut.hashstr27(ut.to_json(subconfig))
#             except KeyError:
#                 print('[deptbl.config] Warning: Config must either'
#                       'contain a string <tablename>_hashid or a dict'
#                       '<tablename>_config')
#                 raise
#     else:
#         config_hashid = 'none'
